# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.util import CLIError

from azure.cli.core.commands import CliCommandType

from ._client_factory import (cf_cdn, cf_custom_domain, cf_endpoints, cf_profiles, cf_origins, cf_resource_usage,
                              cf_edge_nodes, cf_waf_policy, cf_waf_rule_set, cf_origin_groups, cf_afd_endpoints,
                              cf_afd_origins, cf_afd_routes, cf_afd_rule_sets, cf_afd_rules, cf_afd_security_policies,
                              cf_afd_secrets, cf_afd_log_analytics, cf_afd_origin_groups, cf_afd_custom_domain,
                              cf_afd_profiles)


def _not_found(message):
    def _inner_not_found(ex):
        from azure.core.exceptions import ResourceNotFoundError
        if isinstance(ex, ResourceNotFoundError):
            raise CLIError(message)
        raise ex
    return _inner_not_found


def get_custom_sdk(client_factory, exception_handler):
    return CliCommandType(
        operations_tmpl='azure.cli.command_modules.cdn.custom#custom_afdx.{}',
        client_factory=client_factory,
        exception_handler=exception_handler
    )


_not_found_msg = "{}(s) not found. Please verify the resource(s), group or it's parent resources " \
    "exist."


# pylint: disable=too-many-statements
# pylint: disable=too-many-locals
def load_command_table(self, _):
    profile_not_found_msg = _not_found_msg.format('Profile')
    endpoint_not_found_msg = _not_found_msg.format('Endpoint')
    cd_not_found_msg = _not_found_msg.format('Custom Domain')
    origin_not_found_msg = _not_found_msg.format('Origin')
    origin_group_not_found_msg = _not_found_msg.format('Origin Group')
    waf_policy_not_found_msg = _not_found_msg.format('WAF Policy')
    route_not_found_msg = _not_found_msg.format('Route')
    rule_set_not_found_msg = _not_found_msg.format('Rule Set')
    rule_not_found_msg = _not_found_msg.format('Rule')
    security_policy_not_found_msg = _not_found_msg.format('Security Policy')
    secret_not_found_msg = _not_found_msg.format('Secret')
    log_analytic_not_found_msg = _not_found_msg.format('Log Analytic')

    cdn_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn#CdnManagementClient.{}',
        client_factory=cf_cdn
    )

    cdn_endpoints_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#EndpointsOperations.{}',
        client_factory=cf_endpoints,
        exception_handler=_not_found(endpoint_not_found_msg)
    )

    cdn_profiles_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#ProfilesOperations.{}',
        client_factory=cf_profiles,
        exception_handler=_not_found(profile_not_found_msg)
    )

    cdn_domain_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#CustomDomainsOperations.{}',
        client_factory=cf_custom_domain,
        exception_handler=_not_found(cd_not_found_msg)
    )

    cdn_origin_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#OriginsOperations.{}',
        client_factory=cf_origins,
        exception_handler=_not_found(origin_not_found_msg)
    )

    cdn_origin_group_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#OriginGroupsOperations.{}',
        client_factory=cf_origin_groups,
        exception_handler=_not_found(origin_group_not_found_msg)
    )

    cdn_edge_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#EdgeNodesOperations.{}',
        client_factory=cf_edge_nodes
    )

    cdn_usage_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#ResourceUsageOperations.{}',
        client_factory=cf_resource_usage
    )

    cdn_waf_policy_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#PoliciesOperations.{}',
        client_factory=cf_waf_policy,
        exception_handler=_not_found(waf_policy_not_found_msg)
    )

    cdn_afd_endpoints_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#AFDEndpointsOperations.{}',
        client_factory=cf_afd_endpoints,
        exception_handler=_not_found(endpoint_not_found_msg)
    )

    cdn_afd_origin_group_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#AFDOriginGroupsOperations.{}',
        client_factory=cf_afd_origin_groups,
        exception_handler=_not_found(origin_group_not_found_msg)
    )

    cdn_afd_origin_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#AFDOriginsOperations.{}',
        client_factory=cf_afd_origins,
        exception_handler=_not_found(origin_not_found_msg)
    )

    cdn_afd_route_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#RoutesOperations.{}',
        client_factory=cf_afd_routes,
        exception_handler=_not_found(route_not_found_msg)
    )

    cdn_afd_rule_set_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#RuleSetsOperations.{}',
        client_factory=cf_afd_rule_sets,
        exception_handler=_not_found(rule_set_not_found_msg)
    )

    cdn_afd_rule_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#RulesOperations.{}',
        client_factory=cf_afd_rules,
        exception_handler=_not_found(rule_not_found_msg)
    )

    cdn_afd_security_policy_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#SecurityPoliciesOperations.{}',
        client_factory=cf_afd_security_policies,
        exception_handler=_not_found(security_policy_not_found_msg)
    )

    cdn_afd_domain_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#AFDCustomDomainsOperations.{}',
        client_factory=cf_afd_custom_domain,
        exception_handler=_not_found(cd_not_found_msg)
    )

    cdn_afd_secret_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#SecretsOperations.{}',
        client_factory=cf_afd_secrets,
        exception_handler=_not_found(secret_not_found_msg)
    )

    cdn_afd_log_analytic_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.cdn.operations#LogAnalyticsOperations.{}',
        client_factory=cf_afd_log_analytics,
        exception_handler=_not_found(log_analytic_not_found_msg)
    )

    with self.command_group('cdn', cdn_sdk) as g:
        g.custom_command('name-exists', 'check_name_availability', client_factory=cf_cdn)

    with self.command_group('cdn', cdn_usage_sdk) as g:
        g.command('usage', 'list')

    with self.command_group('cdn endpoint', cdn_endpoints_sdk) as g:
        for name in ['start', 'stop', 'delete']:
            g.command(name, f"begin_{name}", supports_no_wait=True)
        g.show_command('show', 'get')
        g.command('list', 'list_by_profile')
        g.custom_command('load', 'load_endpoint_content', client_factory=cf_cdn, supports_no_wait=True)
        g.custom_command('purge', 'purge_endpoint_content', client_factory=cf_cdn, supports_no_wait=True)
        g.custom_command('validate-custom-domain', 'validate_custom_domain', client_factory=cf_cdn)
        g.custom_command('create', 'create_endpoint', client_factory=cf_cdn,
                         doc_string_source='azure.mgmt.cdn.models#Endpoint',
                         supports_no_wait=True)
        g.generic_update_command('update', setter_name='begin_update', setter_arg_name='endpoint_update_properties',
                                 custom_func_name='update_endpoint',
                                 doc_string_source='azure.mgmt.cdn.models#EndpointUpdateParameters',
                                 supports_no_wait=True)

    with self.command_group('cdn endpoint waf policy', cdn_endpoints_sdk, is_preview=True) as g:
        g.custom_show_command('show', 'show_endpoint_waf_policy_link', client_factory=cf_endpoints)
        g.custom_command('set', 'set_endpoint_waf_policy_link', client_factory=cf_endpoints)
        g.custom_command('remove', 'remove_endpoint_waf_policy_link', client_factory=cf_endpoints, confirmation=True)

    with self.command_group('cdn endpoint rule', cdn_endpoints_sdk, is_preview=True) as g:
        g.show_command('show', 'get')
        g.custom_command('add', 'add_rule', client_factory=cf_cdn,
                         doc_string_source='azure.mgmt.cdn.models#Endpoint')
        g.custom_command('remove', 'remove_rule', client_factory=cf_cdn,
                         doc_string_source='azure.mgmt.cdn.models#Endpoint')

    with self.command_group('cdn endpoint rule condition', cdn_endpoints_sdk, is_preview=True) as g:
        g.show_command('show', 'get')
        g.custom_command('add', 'add_condition', client_factory=cf_cdn,
                         doc_string_source='azure.mgmt.cdn.models#Endpoint')
        g.custom_command('remove', 'remove_condition', client_factory=cf_cdn,
                         doc_string_source='azure.mgmt.cdn.models#Endpoint')

    with self.command_group('cdn endpoint rule action', cdn_endpoints_sdk, is_preview=True) as g:
        g.show_command('show', 'get')
        g.custom_command('add', 'add_action', client_factory=cf_cdn,
                         doc_string_source='azure.mgmt.cdn.models#Endpoint')
        g.custom_command('remove', 'remove_action', client_factory=cf_cdn,
                         doc_string_source='azure.mgmt.cdn.models#Endpoint')

    with self.command_group('cdn profile', cdn_profiles_sdk) as g:
        g.custom_show_command('show', 'get_profile', client_factory=cf_cdn)
        g.command('usage', 'list_resource_usage')
        g.custom_command('delete', 'delete_profile', client_factory=cf_cdn)
        g.custom_command('list', 'list_profiles', client_factory=cf_cdn)
        g.custom_command('create', 'create_profile', client_factory=cf_cdn)
        g.generic_update_command('update', setter_name='begin_update', custom_func_name='update_profile',
                                 setter_arg_name='profile_update_parameters',
                                 doc_string_source='azure.mgmt.cdn.models#ProfileUpdateParameters')

    with self.command_group('cdn custom-domain', cdn_domain_sdk) as g:
        g.show_command('show', 'get')
        g.command('delete', 'begin_delete')
        g.command('list', 'list_by_endpoint')
        g.custom_command('create', 'create_custom_domain', client_factory=cf_cdn)
        g.custom_command('enable-https', 'enable_custom_https', client_factory=cf_cdn)
        g.command('disable-https', 'disable_custom_https')

    with self.command_group('cdn origin', cdn_origin_sdk) as g:
        g.show_command('show', 'get')
        g.command('list', 'list_by_endpoint')
        g.custom_command('create', 'create_origin', client_factory=cf_origins, is_preview=True)
        g.custom_command('update', 'update_origin', client_factory=cf_origins)
        g.command('delete', 'begin_delete', confirmation=True)

    with self.command_group('cdn origin-group', cdn_origin_group_sdk, is_preview=True) as g:
        g.show_command('show', 'get')
        g.command('list', 'list_by_endpoint')
        g.custom_command('create', 'create_origin_group', client_factory=cf_origin_groups)
        g.custom_command('update', 'update_origin_group', client_factory=cf_origin_groups)
        g.command('delete', 'begin_delete', confirmation=True)

    with self.command_group('cdn edge-node', cdn_edge_sdk) as g:
        g.command('list', 'list')

    with self.command_group('cdn waf policy', cdn_waf_policy_sdk, is_preview=True) as g:
        g.show_command('show', 'get')
        g.command('list', 'list')
        g.custom_command('set', 'set_waf_policy', client_factory=cf_waf_policy)
        g.command('delete', 'delete', confirmation=True)

    with self.command_group('cdn waf policy managed-rule-set', cdn_waf_policy_sdk, is_preview=True) as g:
        g.custom_command('add', 'add_waf_policy_managed_rule_set', client_factory=cf_waf_policy)
        g.custom_command('remove',
                         'remove_waf_policy_managed_rule_set',
                         client_factory=cf_waf_policy,
                         confirmation=True)
        g.custom_command('list', 'list_waf_policy_managed_rule_sets', client_factory=cf_waf_policy)
        g.custom_show_command('show', 'show_waf_policy_managed_rule_set', client_factory=cf_waf_policy)
        g.custom_command('list-available', 'list_waf_managed_rule_set', client_factory=cf_waf_rule_set)

    with self.command_group('cdn waf policy managed-rule-set rule-group-override',
                            cdn_waf_policy_sdk,
                            is_preview=True) as g:
        g.custom_command('set', 'set_waf_managed_rule_group_override', client_factory=cf_waf_policy)
        g.custom_command('delete',
                         'delete_waf_managed_rule_group_override',
                         client_factory=cf_waf_policy,
                         confirmation=True)
        g.custom_command('list', 'list_waf_policy_managed_rule_group_overrides', client_factory=cf_waf_policy)
        g.custom_show_command('show', 'show_waf_managed_rule_group_override', client_factory=cf_waf_policy)
        g.custom_command('list-available', 'list_waf_managed_rule_groups', client_factory=cf_waf_rule_set)

    with self.command_group('cdn waf policy custom-rule', cdn_waf_policy_sdk, is_preview=True) as g:
        g.custom_command('set', 'set_waf_custom_rule', client_factory=cf_waf_policy)
        g.custom_command('delete', 'delete_waf_custom_rule', client_factory=cf_waf_policy, confirmation=True)
        g.custom_command('list', 'list_waf_custom_rules', client_factory=cf_waf_policy)
        g.custom_show_command('show', 'show_waf_custom_rule', client_factory=cf_waf_policy)

    with self.command_group('cdn waf policy rate-limit-rule', cdn_waf_policy_sdk, is_preview=True) as g:
        g.custom_command('set', 'set_waf_rate_limit_rule', client_factory=cf_waf_policy)
        g.custom_command('delete', 'delete_waf_rate_limit_rule', client_factory=cf_waf_policy, confirmation=True)
        g.custom_command('list', 'list_waf_rate_limit_rules', client_factory=cf_waf_policy)
        g.custom_show_command('show', 'show_waf_rate_limit_rule', client_factory=cf_waf_policy)

    with self.command_group('afd', is_preview=True):
        pass

    with self.command_group('afd profile', cdn_profiles_sdk,
                            custom_command_type=get_custom_sdk(cf_profiles, _not_found(profile_not_found_msg))) as g:
        g.custom_show_command('show', 'get_afd_profile')
        g.custom_command('delete', 'delete_afd_profile')
        g.custom_command('update', 'update_afd_profile')
        g.custom_command('list', 'list_afd_profiles')
        g.custom_command('create', 'create_afd_profile')
        g.custom_command('usage', 'list_resource_usage', client_factory=cf_afd_profiles)

    with self.command_group('afd endpoint', cdn_afd_endpoints_sdk,
                            client_factory=cf_afd_endpoints) as g:
        g.show_command('show', 'get')
        g.command('list', 'list_by_profile')
        g.custom_command('purge', 'purge_afd_endpoint_content', supports_no_wait=True)
        g.command('delete', 'begin_delete', confirmation=True)

        g.custom_command('update', 'update_afd_endpoint')
        g.custom_command('create', 'create_afd_endpoint',
                         doc_string_source='azure.mgmt.cdn.models#AFDEndpoint')

    with self.command_group('afd origin-group', cdn_afd_origin_group_sdk,
                            client_factory=cf_afd_origin_groups) as g:
        g.show_command('show', 'get')
        g.command('list', 'list_by_profile')
        g.custom_command('create', 'create_afd_origin_group')
        g.custom_command('update', 'update_afd_origin_group')
        g.command('delete', 'begin_delete', confirmation=True)

    with self.command_group('afd origin', cdn_afd_origin_sdk,
                            client_factory=cf_afd_origins) as g:
        g.show_command('show', 'get')
        g.command('list', 'list_by_origin_group')
        g.custom_command('create', 'create_afd_origin')
        g.custom_command('update', 'update_afd_origin')
        g.command('delete', 'begin_delete', confirmation=True)

    with self.command_group('afd route', cdn_afd_route_sdk,
                            client_factory=cf_afd_routes) as g:
        g.show_command('show', 'get')
        g.command('list', 'list_by_endpoint')
        g.custom_command('create', 'create_afd_route')
        g.custom_command('update', 'update_afd_route')
        g.command('delete', 'begin_delete', confirmation=True)

    with self.command_group('afd rule-set', cdn_afd_rule_set_sdk,
                            client_factory=cf_afd_rule_sets) as g:
        g.show_command('show', 'get')
        g.command('list', 'list_by_profile')
        g.custom_command('create', 'create_afd_rule_set')
        g.command('delete', 'begin_delete', confirmation=True)

    with self.command_group('afd rule', cdn_afd_rule_sdk,
                            client_factory=cf_afd_rules) as g:
        g.show_command('show', 'get')
        g.command('list', 'list_by_rule_set')
        g.custom_command('create', 'create_afd_rule')
        g.command('delete', 'begin_delete', confirmation=True)

    with self.command_group('afd rule condition',
                            cdn_afd_rule_sdk,
                            custom_command_type=get_custom_sdk(cf_afd_rules, _not_found(rule_not_found_msg))) as g:
        g.custom_command('add', 'add_afd_rule_condition',
                         doc_string_source='azure.mgmt.cdn.models#Rule')
        g.custom_command('remove', 'remove_afd_rule_condition',
                         doc_string_source='azure.mgmt.cdn.models#Rule')
        g.custom_command('list', 'list_afd_rule_condition')

    with self.command_group('afd rule action',
                            cdn_afd_rule_sdk,
                            custom_command_type=get_custom_sdk(cf_afd_rules, _not_found(rule_not_found_msg))) as g:
        g.custom_command('add', 'add_afd_rule_action',
                         doc_string_source='azure.mgmt.cdn.models#Rule')
        g.custom_command('remove', 'remove_afd_rule_action',
                         doc_string_source='azure.mgmt.cdn.models#Rule')
        g.custom_command('list', 'list_afd_rule_action')

    with self.command_group('afd security-policy', cdn_afd_security_policy_sdk,
                            client_factory=cf_afd_security_policies) as g:
        g.show_command('show', 'get')
        g.command('list', 'list_by_profile')
        g.custom_command('create', 'create_afd_security_policy')
        g.custom_command('update', 'update_afd_security_policy')
        g.command('delete', 'begin_delete', confirmation=True)

    with self.command_group('afd custom-domain', cdn_afd_domain_sdk,
                            client_factory=cf_afd_custom_domain) as g:
        g.show_command('show', 'get')
        g.wait_command('wait')
        g.command('delete', 'begin_delete', confirmation=True)
        g.command('list', 'list_by_profile')
        g.custom_command('create', 'create_afd_custom_domain',
                         supports_no_wait=True)
        g.custom_command('update', 'update_afd_custom_domain')

    with self.command_group('afd secret', cdn_afd_secret_sdk,
                            client_factory=cf_afd_secrets) as g:
        g.show_command('show', 'get')
        g.command('delete', 'begin_delete', confirmation=True)
        g.command('list', 'list_by_profile')
        g.custom_command('create', 'create_afd_secret')
        g.custom_command('update', 'update_afd_secret')

    with self.command_group('afd log-analytic metric', cdn_afd_log_analytic_sdk) as g:
        g.command('list', 'get_log_analytics_metrics')

    with self.command_group('afd log-analytic ranking', cdn_afd_log_analytic_sdk) as g:
        g.command('list', 'get_log_analytics_rankings')

    with self.command_group('afd log-analytic location', cdn_afd_log_analytic_sdk) as g:
        g.command('list', 'get_log_analytics_locations')

    with self.command_group('afd log-analytic resource', cdn_afd_log_analytic_sdk) as g:
        g.command('list', 'get_log_analytics_resources')

    with self.command_group('afd waf-log-analytic metric', cdn_afd_log_analytic_sdk) as g:
        g.command('list', 'get_waf_log_analytics_metrics')

    with self.command_group('afd waf-log-analytic ranking', cdn_afd_log_analytic_sdk) as g:
        g.command('list', 'get_waf_log_analytics_rankings')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------


def network_client_factory(cli_ctx, **kwargs):
    from azure.cli.core.profiles import ResourceType
    from azure.cli.core.commands.client_factory import get_mgmt_service_client
    return get_mgmt_service_client(cli_ctx, ResourceType.MGMT_NETWORK, **kwargs)


def cf_nat_gateways(cli_ctx, _):
    return network_client_factory(cli_ctx).nat_gateways



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

def file_related_exception_handler(ex):
    from azure.cli.core.azclierror import FileOperationError
    if isinstance(ex, FileNotFoundError):
        raise FileOperationError(ex, recommendation='Please check the file path.')
    if isinstance(ex, PermissionError):
        raise FileOperationError(ex,
                                 recommendation='Please make sure you have enough permissions on the file/directory.')
    if isinstance(ex, IsADirectoryError):
        raise FileOperationError(ex, recommendation='File is expected, not a directory.')
    if isinstance(ex, NotADirectoryError):
        raise FileOperationError(ex, recommendation='Directory is expected, not a file.')
    import sys
    from six import reraise
    reraise(*sys.exc_info())



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.command_modules.vm._client_factory import (cf_vm, cf_avail_set, cf_ni,
                                                          cf_vm_ext, cf_vm_ext_image,
                                                          cf_vm_image, cf_vm_image_term, cf_usage,
                                                          cf_vmss, cf_vmss_vm,
                                                          cf_vm_sizes, cf_disks, cf_snapshots,
                                                          cf_disk_accesses, cf_images, cf_run_commands,
                                                          cf_rolling_upgrade_commands, cf_galleries,
                                                          cf_gallery_images, cf_gallery_image_versions,
                                                          cf_proximity_placement_groups,
                                                          cf_dedicated_hosts, cf_dedicated_host_groups,
                                                          cf_log_analytics_data_plane,
                                                          cf_disk_encryption_set)
from azure.cli.command_modules.vm._format import (
    transform_ip_addresses, transform_vm, transform_vm_create_output, transform_vm_usage_list, transform_vm_list,
    transform_sku_for_table_output, transform_disk_show_table_output, transform_extension_show_table_output,
    get_vmss_table_output_transformer, transform_vm_encryption_show_table_output, transform_log_analytics_query_output)
from azure.cli.command_modules.vm._validators import (
    process_vm_create_namespace, process_vmss_create_namespace, process_image_create_namespace,
    process_disk_or_snapshot_create_namespace, process_disk_encryption_namespace, process_assign_identity_namespace,
    process_remove_identity_namespace, process_vm_secret_format, process_vm_vmss_stop, validate_vmss_update_namespace)

from azure.cli.command_modules.vm._image_builder import (
    process_image_template_create_namespace, process_img_tmpl_output_add_namespace,
    process_img_tmpl_customizer_add_namespace, image_builder_client_factory, cf_img_bldr_image_templates)

from azure.cli.core.commands import DeploymentOutputLongRunningOperation, CliCommandType
from azure.cli.core.commands.arm import deployment_validate_table_format, handle_template_based_exception

from azure.cli.command_modules.monitor._exception_handler import exception_handler as monitor_exception_handler
from azure.cli.command_modules.monitor._client_factory import cf_metric_def
from azure.cli.core.profiles import ResourceType


# pylint: disable=line-too-long, too-many-statements, too-many-locals
def load_command_table(self, _):

    custom_tmpl = 'azure.cli.command_modules.vm.custom#{}'

    compute_custom = CliCommandType(operations_tmpl=custom_tmpl)

    compute_disk_encryption_custom = CliCommandType(
        operations_tmpl='azure.cli.command_modules.vm.disk_encryption#{}',
        operation_group='virtual_machines'
    )

    image_builder_custom = CliCommandType(
        operations_tmpl='azure.cli.command_modules.vm._image_builder#{}',
        client_factory=image_builder_client_factory
    )

    compute_availset_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#AvailabilitySetsOperations.{}',
        client_factory=cf_avail_set,
        operation_group='availability_sets'
    )

    compute_disk_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#DisksOperations.{}',
        client_factory=cf_disks,
        operation_group='disks'
    )

    compute_disk_access_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#DiskAccessesOperations.{}',
        client_factory=cf_disk_accesses,
        operation_group='disk_accesses'
    )

    compute_image_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#ImagesOperations.{}',
        client_factory=cf_images
    )

    compute_snapshot_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#SnapshotsOperations.{}',
        client_factory=cf_snapshots
    )

    compute_vm_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#VirtualMachinesOperations.{}',
        client_factory=cf_vm
    )

    compute_vm_extension_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#VirtualMachineExtensionsOperations.{}',
        client_factory=cf_vm_ext
    )

    compute_vm_extension_image_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#VirtualMachineExtensionImagesOperations.{}',
        client_factory=cf_vm_ext_image
    )

    compute_vm_image_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#VirtualMachineImagesOperations.{}',
        client_factory=cf_vm_image
    )

    compute_vm_image_term_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.marketplaceordering.operations#MarketplaceAgreementsOperations.{}',
        client_factory=cf_vm_image_term
    )

    compute_vm_usage_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#UsageOperations.{}',
        client_factory=cf_usage
    )

    compute_vm_run_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#VirtualMachineRunCommandsOperations.{}',
        client_factory=cf_run_commands
    )

    compute_vm_size_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#VirtualMachineSizesOperations.{}',
        client_factory=cf_vm_sizes
    )

    compute_vmss_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#VirtualMachineScaleSetsOperations.{}',
        client_factory=cf_vmss,
        operation_group='virtual_machine_scale_sets'
    )

    compute_vmss_rolling_upgrade_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#VirtualMachineScaleSetRollingUpgradesOperations.{}',
        client_factory=cf_rolling_upgrade_commands,
        operation_group='virtual_machine_scale_sets'
    )

    compute_vmss_vm_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#VirtualMachineScaleSetVMsOperations.{}',
        client_factory=cf_vmss_vm,
        operation_group='virtual_machine_scale_sets'
    )

    network_nic_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.network.operations#NetworkInterfacesOperations.{}',
        client_factory=cf_ni
    )

    compute_galleries_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#GalleriesOperations.{}',
        client_factory=cf_galleries,
    )

    compute_gallery_images_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#GalleryImagesOperations.{}',
        client_factory=cf_gallery_images,
    )

    compute_gallery_image_versions_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#GalleryImageVersionsOperations.{}',
        client_factory=cf_gallery_image_versions,
    )

    compute_proximity_placement_groups_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#ProximityPlacementGroupsOperations.{}',
    )

    compute_dedicated_host_sdk = CliCommandType(
        operations_tmpl="azure.mgmt.compute.operations#DedicatedHostsOperations.{}",
        client_factory=cf_dedicated_hosts,
    )

    compute_dedicated_host_groups_sdk = CliCommandType(
        operations_tmpl="azure.mgmt.compute.operations#DedicatedHostGroupsOperations.{}",
        client_factory=cf_dedicated_host_groups,
    )

    image_builder_image_templates_sdk = CliCommandType(
        operations_tmpl="azure.mgmt.imagebuilder.operations#VirtualMachineImageTemplatesOperations.{}",
        client_factory=cf_img_bldr_image_templates,
    )

    compute_disk_encryption_set_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.compute.operations#DiskEncryptionSetsOperations.{}',
        client_factory=cf_disk_encryption_set
    )

    monitor_custom = CliCommandType(
        operations_tmpl='azure.cli.command_modules.monitor.custom#{}',
        exception_handler=monitor_exception_handler
    )

    metric_definitions_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.monitor.operations#MetricDefinitionsOperations.{}',
        resource_type=ResourceType.MGMT_MONITOR,
        client_factory=cf_metric_def,
        operation_group='metric_definitions',
        exception_handler=monitor_exception_handler
    )

    with self.command_group('disk', compute_disk_sdk, operation_group='disks', min_api='2017-03-30') as g:
        g.custom_command('create', 'create_managed_disk', supports_no_wait=True, table_transformer=transform_disk_show_table_output, validator=process_disk_or_snapshot_create_namespace)
        g.command('delete', 'begin_delete', supports_no_wait=True, confirmation=True)
        g.custom_command('grant-access', 'grant_disk_access')
        g.custom_command('list', 'list_managed_disks', table_transformer='[].' + transform_disk_show_table_output)
        g.command('revoke-access', 'begin_revoke_access')
        g.show_command('show', 'get', table_transformer=transform_disk_show_table_output)
        g.generic_update_command('update', custom_func_name='update_managed_disk', setter_name='begin_create_or_update', setter_arg_name='disk', supports_no_wait=True)
        g.wait_command('wait')

    with self.command_group('disk-encryption-set', compute_disk_encryption_set_sdk, operation_group='disk_encryption_sets', client_factory=cf_disk_encryption_set, min_api='2019-07-01') as g:
        g.custom_command('create', 'create_disk_encryption_set', supports_no_wait=True)
        g.command('delete', 'begin_delete')
        g.generic_update_command('update', custom_func_name='update_disk_encryption_set', setter_arg_name='disk_encryption_set', setter_name='begin_create_or_update')
        g.show_command('show', 'get')
        g.custom_command('list', 'list_disk_encryption_sets')
        g.command('list-associated-resources', 'list_associated_resources', min_api='2020-06-30')

    with self.command_group('disk-access', compute_disk_access_sdk, operation_group='disk_accesses', client_factory=cf_disk_accesses, min_api='2020-05-01') as g:
        g.custom_command('create', 'create_disk_access', supports_no_wait=True)
        g.generic_update_command('update', setter_name='set_disk_access', setter_type=compute_custom, supports_no_wait=True)
        g.show_command('show', 'get')
        g.custom_command('list', 'list_disk_accesses')
        g.wait_command('wait')
        g.command('delete', 'begin_delete')

    with self.command_group('image', compute_image_sdk, min_api='2016-04-30-preview') as g:
        g.custom_command('create', 'create_image', validator=process_image_create_namespace)
        g.custom_command('list', 'list_images')
        g.show_command('show', 'get')
        g.command('delete', 'begin_delete')
        g.generic_update_command('update', setter_name='begin_create_or_update', custom_func_name='update_image')

    with self.command_group('image builder', image_builder_image_templates_sdk, custom_command_type=image_builder_custom) as g:
        g.custom_command('create', 'create_image_template', supports_no_wait=True, supports_local_cache=True, validator=process_image_template_create_namespace)
        g.custom_command('list', 'list_image_templates')
        g.show_command('show', 'get')
        g.command('delete', 'delete')
        g.generic_update_command('update', 'create_or_update', supports_local_cache=True)  # todo Update fails for now as service does not support updates
        g.wait_command('wait')
        g.command('run', 'run', supports_no_wait=True)
        g.custom_command('show-runs', 'show_build_output')
        g.command('cancel', 'cancel')

    with self.command_group('image builder customizer', image_builder_image_templates_sdk, custom_command_type=image_builder_custom) as g:
        g.custom_command('add', 'add_template_customizer', supports_local_cache=True, validator=process_img_tmpl_customizer_add_namespace)
        g.custom_command('remove', 'remove_template_customizer', supports_local_cache=True)
        g.custom_command('clear', 'clear_template_customizer', supports_local_cache=True)

    with self.command_group('image builder output', image_builder_image_templates_sdk, custom_command_type=image_builder_custom) as g:
        g.custom_command('add', 'add_template_output', supports_local_cache=True, validator=process_img_tmpl_output_add_namespace)
        g.custom_command('remove', 'remove_template_output', supports_local_cache=True)
        g.custom_command('clear', 'clear_template_output', supports_local_cache=True)

    with self.command_group('snapshot', compute_snapshot_sdk, operation_group='snapshots', min_api='2016-04-30-preview') as g:
        g.custom_command('create', 'create_snapshot', validator=process_disk_or_snapshot_create_namespace, supports_no_wait=True)
        g.command('delete', 'begin_delete')
        g.custom_command('grant-access', 'grant_snapshot_access')
        g.custom_command('list', 'list_snapshots')
        g.command('revoke-access', 'begin_revoke_access')
        g.show_command('show', 'get')
        g.generic_update_command('update', custom_func_name='update_snapshot', setter_name='begin_create_or_update', setter_arg_name='snapshot', supports_no_wait=True)
        g.wait_command('wait')

    with self.command_group('vm', compute_vm_sdk) as g:
        g.custom_command('identity assign', 'assign_vm_identity', validator=process_assign_identity_namespace)
        g.custom_command('identity remove', 'remove_vm_identity', validator=process_remove_identity_namespace, min_api='2017-12-01')
        g.custom_show_command('identity show', 'show_vm_identity')

        g.custom_command('capture', 'capture_vm')
        g.custom_command('create', 'create_vm', transform=transform_vm_create_output, supports_no_wait=True, table_transformer=deployment_validate_table_format, validator=process_vm_create_namespace, exception_handler=handle_template_based_exception)
        g.command('convert', 'begin_convert_to_managed_disks', min_api='2016-04-30-preview')
        g.command('deallocate', 'begin_deallocate', supports_no_wait=True)
        g.command('delete', 'begin_delete', confirmation=True, supports_no_wait=True)
        g.command('generalize', 'generalize', supports_no_wait=True)
        g.custom_command('get-instance-view', 'get_instance_view', table_transformer='{Name:name, ResourceGroup:resourceGroup, Location:location, ProvisioningState:provisioningState, PowerState:instanceView.statuses[1].displayStatus}')
        g.custom_command('list', 'list_vm', table_transformer=transform_vm_list)
        g.custom_command('list-ip-addresses', 'list_vm_ip_addresses', table_transformer=transform_ip_addresses)
        g.command('list-sizes', 'list', command_type=compute_vm_size_sdk)
        g.custom_command('list-skus', 'list_skus', table_transformer=transform_sku_for_table_output, min_api='2017-03-30')
        g.command('list-usage', 'list', command_type=compute_vm_usage_sdk, transform=transform_vm_usage_list, table_transformer='[].{Name:localName, CurrentValue:currentValue, Limit:limit}')
        g.command('list-vm-resize-options', 'list_available_sizes')
        g.custom_command('open-port', 'open_vm_port')
        g.command('perform-maintenance', 'begin_perform_maintenance', min_api='2017-03-30')
        g.command('redeploy', 'begin_redeploy', supports_no_wait=True)
        g.custom_command('resize', 'resize_vm', supports_no_wait=True)
        g.custom_command('restart', 'restart_vm', supports_no_wait=True)
        g.custom_show_command('show', 'show_vm', table_transformer=transform_vm)
        g.command('simulate-eviction', 'simulate_eviction', min_api='2019-12-01')
        g.command('start', 'begin_start', supports_no_wait=True)
        g.command('stop', 'begin_power_off', supports_no_wait=True, validator=process_vm_vmss_stop)
        g.command('reapply', 'begin_reapply', supports_no_wait=True, min_api='2019-07-01')
        g.generic_update_command('update', getter_name='get_vm_to_update', setter_name='update_vm', setter_type=compute_custom, command_type=compute_custom, supports_no_wait=True)
        g.wait_command('wait', getter_name='get_instance_view', getter_type=compute_custom)
        g.custom_command('auto-shutdown', 'auto_shutdown_vm')
        g.command('assess-patches', 'begin_assess_patches', min_api='2020-06-01')

    with self.command_group('vm', compute_vm_sdk, client_factory=cf_vm) as g:
        g.custom_command('install-patches', 'install_vm_patches', supports_no_wait=True, min_api='2020-12-01')

    with self.command_group('vm availability-set', compute_availset_sdk) as g:
        g.custom_command('convert', 'convert_av_set_to_managed_disk', min_api='2016-04-30-preview')
        g.custom_command('create', 'create_av_set', table_transformer=deployment_validate_table_format, supports_no_wait=True, exception_handler=handle_template_based_exception)
        g.command('delete', 'delete')
        g.custom_command('list', 'list_av_sets')
        g.command('list-sizes', 'list_available_sizes')
        g.show_command('show', 'get')
        g.generic_update_command('update', custom_func_name='update_av_set')

    with self.command_group('vm boot-diagnostics', compute_vm_sdk) as g:
        g.custom_command('disable', 'disable_boot_diagnostics')
        g.custom_command('enable', 'enable_boot_diagnostics')
        g.custom_command('get-boot-log', 'get_boot_log')
        g.custom_command('get-boot-log-uris', 'get_boot_log_uris', min_api='2020-06-01')

    with self.command_group('vm diagnostics', compute_vm_sdk) as g:
        g.custom_command('set', 'set_diagnostics_extension')
        g.custom_command('get-default-config', 'show_default_diagnostics_configuration')

    with self.command_group('vm disk', compute_vm_sdk, min_api='2017-03-30') as g:
        g.custom_command('attach', 'attach_managed_data_disk')
        g.custom_command('detach', 'detach_data_disk')

    with self.command_group('vm encryption', custom_command_type=compute_disk_encryption_custom) as g:
        g.custom_command('enable', 'encrypt_vm', validator=process_disk_encryption_namespace)
        g.custom_command('disable', 'decrypt_vm')
        g.custom_show_command('show', 'show_vm_encryption_status', table_transformer=transform_vm_encryption_show_table_output)

    with self.command_group('vm extension', compute_vm_extension_sdk) as g:
        g.command('delete', 'begin_delete', supports_no_wait=True)
        g.show_command('show', 'get', table_transformer=transform_extension_show_table_output)
        g.custom_command('set', 'set_extension', supports_no_wait=True)
        g.custom_command('list', 'list_extensions', table_transformer='[].' + transform_extension_show_table_output)
        g.wait_command('wait')

    with self.command_group('vm extension image', compute_vm_extension_image_sdk) as g:
        g.show_command('show', 'get')
        g.command('list-names', 'list_types')
        g.command('list-versions', 'list_versions')
        g.custom_command('list', 'list_vm_extension_images')

    with self.command_group('vm image', compute_vm_image_sdk) as g:
        g.command('list-offers', 'list_offers')
        g.command('list-publishers', 'list_publishers')
        g.command('list-skus', 'list_skus')
        g.custom_command('list', 'list_vm_images')
        g.custom_command('accept-terms', 'accept_market_ordering_terms',
                         deprecate_info=g.deprecate(redirect='az vm image terms accept', expiration='3.0.0'))
        g.custom_show_command('show', 'show_vm_image')

    with self.command_group('vm image terms', compute_vm_image_term_sdk, validator=None) as g:
        g.custom_command('accept', 'accept_terms')
        g.custom_command('cancel', 'cancel_terms')
        g.custom_show_command('show', 'get_terms')

    with self.command_group('vm nic', compute_vm_sdk) as g:
        g.custom_command('add', 'add_vm_nic')
        g.custom_command('remove', 'remove_vm_nic')
        g.custom_command('set', 'set_vm_nic')
        g.custom_show_command('show', 'show_vm_nic')
        g.custom_command('list', 'list_vm_nics')

    with self.command_group('vm run-command', compute_vm_run_sdk, operation_group='virtual_machine_run_commands', min_api='2017-03-30') as g:
        g.custom_command('invoke', 'vm_run_command_invoke')
        g.command('list', 'list')
        g.show_command('show', 'get')

    with self.command_group('vm secret', compute_vm_sdk) as g:
        g.custom_command('format', 'get_vm_format_secret', validator=process_vm_secret_format)
        g.custom_command('add', 'add_vm_secret')
        g.custom_command('list', 'list_vm_secrets')
        g.custom_command('remove', 'remove_vm_secret')

    with self.command_group('vm unmanaged-disk', compute_vm_sdk) as g:
        g.custom_command('attach', 'attach_unmanaged_data_disk')
        g.custom_command('detach', 'detach_data_disk')
        g.custom_command('list', 'list_unmanaged_disks')

    with self.command_group('vm user', compute_vm_sdk, supports_no_wait=True) as g:
        g.custom_command('update', 'set_user')
        g.custom_command('delete', 'delete_user')
        g.custom_command('reset-ssh', 'reset_linux_ssh')

    with self.command_group('vm host', compute_dedicated_host_sdk, client_factory=cf_dedicated_hosts,
                            min_api='2019-03-01') as g:
        g.show_command('show', 'get')
        g.custom_command('get-instance-view', 'get_dedicated_host_instance_view')
        g.custom_command('create', 'create_dedicated_host')
        g.command('list', 'list_by_host_group')
        g.generic_update_command('update', setter_name='begin_create_or_update')
        g.command('delete', 'begin_delete', confirmation=True)

    with self.command_group('vm host group', compute_dedicated_host_groups_sdk, client_factory=cf_dedicated_host_groups,
                            min_api='2019-03-01') as g:
        g.show_command('show', 'get')
        g.custom_command('get-instance-view', 'get_dedicated_host_group_instance_view', min_api='2020-06-01')
        g.custom_command('create', 'create_dedicated_host_group')
        g.custom_command('list', 'list_dedicated_host_groups')
        g.generic_update_command('update')
        g.command('delete', 'delete', confirmation=True)

    with self.command_group('vmss', compute_vmss_sdk, operation_group='virtual_machine_scale_sets') as g:
        g.custom_command('identity assign', 'assign_vmss_identity', validator=process_assign_identity_namespace)
        g.custom_command('identity remove', 'remove_vmss_identity', validator=process_remove_identity_namespace, min_api='2017-12-01', is_preview=True)
        g.custom_show_command('identity show', 'show_vmss_identity')
        g.custom_command('create', 'create_vmss', transform=DeploymentOutputLongRunningOperation(self.cli_ctx, 'Starting vmss create'), supports_no_wait=True, table_transformer=deployment_validate_table_format, validator=process_vmss_create_namespace, exception_handler=handle_template_based_exception)
        g.custom_command('deallocate', 'deallocate_vmss', supports_no_wait=True)
        g.command('delete', 'begin_delete', supports_no_wait=True)
        g.custom_command('delete-instances', 'delete_vmss_instances', supports_no_wait=True)
        g.custom_command('get-instance-view', 'get_vmss_instance_view', table_transformer='{ProvisioningState:statuses[0].displayStatus, PowerState:statuses[1].displayStatus}')
        g.custom_command('list', 'list_vmss', table_transformer=get_vmss_table_output_transformer(self))
        g.command('list-instances', 'list', command_type=compute_vmss_vm_sdk)
        g.custom_command('list-instance-connection-info', 'list_vmss_instance_connection_info')
        g.custom_command('list-instance-public-ips', 'list_vmss_instance_public_ips')
        g.command('list-skus', 'list_skus')
        g.custom_command('reimage', 'reimage_vmss', supports_no_wait=True, min_api='2017-03-30')
        g.command('perform-maintenance', 'begin_perform_maintenance', min_api='2017-12-01')
        g.custom_command('restart', 'restart_vmss', supports_no_wait=True)
        g.custom_command('scale', 'scale_vmss', supports_no_wait=True)
        g.custom_show_command('show', 'get_vmss', table_transformer=get_vmss_table_output_transformer(self, False))
        g.command('simulate-eviction', 'simulate_eviction', command_type=compute_vmss_vm_sdk, min_api='2019-12-01')
        g.custom_command('start', 'start_vmss', supports_no_wait=True)
        g.custom_command('stop', 'stop_vmss', supports_no_wait=True, validator=process_vm_vmss_stop)
        g.generic_update_command('update', getter_name='get_vmss_modified', setter_name='update_vmss', supports_no_wait=True, command_type=compute_custom, validator=validate_vmss_update_namespace)
        g.custom_command('update-instances', 'update_vmss_instances', supports_no_wait=True)
        g.wait_command('wait', getter_name='get_vmss', getter_type=compute_custom)
        g.command('get-os-upgrade-history', 'get_os_upgrade_history', min_api='2018-10-01')
        g.custom_command('set-orchestration-service-state', 'set_orchestration_service_state', supports_no_wait=True)

    with self.command_group('vmss diagnostics', compute_vmss_sdk) as g:
        g.custom_command('set', 'set_vmss_diagnostics_extension')
        g.custom_command('get-default-config', 'show_default_diagnostics_configuration')

    with self.command_group('vmss disk', compute_vmss_sdk, min_api='2017-03-30') as g:
        g.custom_command('attach', 'attach_managed_data_disk_to_vmss')
        g.custom_command('detach', 'detach_disk_from_vmss')

    with self.command_group('vmss encryption', custom_command_type=compute_disk_encryption_custom, min_api='2017-03-30') as g:
        g.custom_command('enable', 'encrypt_vmss', validator=process_disk_encryption_namespace)
        g.custom_command('disable', 'decrypt_vmss')
        g.custom_show_command('show', 'show_vmss_encryption_status')

    with self.command_group('vmss extension', compute_vmss_sdk) as g:
        g.custom_command('delete', 'delete_vmss_extension', supports_no_wait=True)
        g.custom_show_command('show', 'get_vmss_extension')
        g.custom_command('set', 'set_vmss_extension', supports_no_wait=True)
        g.custom_command('list', 'list_vmss_extensions')
        g.custom_command('upgrade', 'upgrade_vmss_extension', min_api='2020-06-01', supports_no_wait=True)

    with self.command_group('vmss extension image', compute_vm_extension_image_sdk) as g:
        g.show_command('show', 'get')
        g.command('list-names', 'list_types')
        g.command('list-versions', 'list_versions')
        g.custom_command('list', 'list_vm_extension_images')

    with self.command_group('vmss nic', network_nic_sdk) as g:
        g.command('list', 'list_virtual_machine_scale_set_network_interfaces')
        g.command('list-vm-nics', 'list_virtual_machine_scale_set_vm_network_interfaces')
        g.show_command('show', 'get_virtual_machine_scale_set_network_interface')

    with self.command_group('vmss run-command', compute_vm_run_sdk, min_api='2018-04-01') as g:
        g.custom_command('invoke', 'vmss_run_command_invoke')
        g.command('list', 'list')
        g.show_command('show', 'get')

    with self.command_group('vmss rolling-upgrade', compute_vmss_rolling_upgrade_sdk, min_api='2017-03-30') as g:
        g.command('cancel', 'begin_cancel')
        g.command('get-latest', 'get_latest')
        g.command('start', 'begin_start_os_upgrade')

    with self.command_group('sig', compute_galleries_sdk, operation_group='galleries', min_api='2018-06-01') as g:
        g.custom_command('create', 'create_image_gallery')
        g.show_command('show', 'get')
        g.custom_command('list', 'list_image_galleries')
        g.command('delete', 'begin_delete')
        g.generic_update_command('update', setter_name='begin_create_or_update', setter_arg_name='gallery')

    with self.command_group('sig image-definition', compute_gallery_images_sdk, operation_group='gallery_images', min_api='2018-06-01') as g:
        g.custom_command('create', 'create_gallery_image')
        g.command('list', 'list_by_gallery')
        g.show_command('show', 'get')
        g.command('delete', 'begin_delete')
        g.generic_update_command('update', setter_name='begin_create_or_update', setter_arg_name='gallery_image')

    with self.command_group('sig image-version', compute_gallery_image_versions_sdk, operation_group='gallery_image_versions', min_api='2018-06-01') as g:
        g.command('delete', 'begin_delete')
        g.show_command('show', 'get', table_transformer='{Name:name, ResourceGroup:resourceGroup, ProvisioningState:provisioningState, TargetRegions: publishingProfile.targetRegions && join(`, `, publishingProfile.targetRegions[*].name), ReplicationState:replicationStatus.aggregatedState}')
        g.command('list', 'list_by_gallery_image')
        g.custom_command('create', 'create_image_version', supports_no_wait=True)
        g.generic_update_command('update', getter_name='get_image_version_to_update', setter_arg_name='gallery_image_version', setter_name='update_image_version', setter_type=compute_custom, command_type=compute_custom, supports_no_wait=True)
        g.wait_command('wait')

    with self.command_group('ppg', compute_proximity_placement_groups_sdk, min_api='2018-04-01', client_factory=cf_proximity_placement_groups) as g:
        g.show_command('show', 'get')
        g.custom_command('create', 'create_proximity_placement_group')
        g.custom_command('list', 'list_proximity_placement_groups')
        g.generic_update_command('update')
        g.command('delete', 'delete')

    with self.command_group('vm monitor log', client_factory=cf_log_analytics_data_plane) as g:
        g.custom_command('show', 'execute_query_for_vm', transform=transform_log_analytics_query_output)  # pylint: disable=show-command

    with self.command_group('vm monitor metrics', custom_command_type=monitor_custom, command_type=metric_definitions_sdk, resource_type=ResourceType.MGMT_MONITOR, operation_group='metric_definitions', min_api='2018-01-01', is_preview=True) as g:
        from azure.cli.command_modules.monitor.transformers import metrics_table, metrics_definitions_table
        from azure.cli.core.profiles._shared import APIVersionException
        try:
            g.custom_command('tail', 'list_metrics', command_type=monitor_custom, table_transformer=metrics_table)
            g.command('list-definitions', 'list', table_transformer=metrics_definitions_table)
        except APIVersionException:
            pass



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
from argcomplete.completers import FilesCompleter

from knack.arguments import CLIArgumentType

from azure.cli.core.commands.parameters import (
    tags_type, get_resource_name_completion_list, resource_group_name_type, get_enum_type)

from azure.cli.command_modules.dla._validators import validate_resource_group_name, datetime_format


# pylint: disable=line-too-long, too-many-statements
def load_arguments(self, _):
    from azure.mgmt.datalake.analytics.account.models import (FirewallState, TierType, FirewallAllowAzureIpsState,
                                                              AADObjectType)

    from azure.mgmt.datalake.analytics.job.models import (CompileMode, JobState, JobResult)

    datalake_analytics_name_type = CLIArgumentType(help='Name of the Data Lake Analytics account.', options_list=('--account_name',), completer=get_resource_name_completion_list('Microsoft.DataLakeAnalytics/accounts'), id_part='name')

    # PARAMETER REGISTRATIONS
    # common
    with self.argument_context('dla') as c:
        c.argument('resource_group_name', resource_group_name_type, id_part=None, required=False, help='If not specified, will attempt to discover the resource group for the specified Data Lake Analytics account.', validator=validate_resource_group_name)
        c.argument('top', help='Maximum number of items to return.', type=int)
        c.argument('skip', help='The number of items to skip over before returning elements.', type=int)
        c.argument('count', help='The Boolean value of true or false to request a count of the matching resources included with the resources in the response, e.g. Categories?$count=true.', type=bool)
        c.argument('account_name', datalake_analytics_name_type, options_list=['--account', '-n'])

    # account
    with self.argument_context('dla account') as c:
        c.argument('tags', tags_type)
        c.argument('tier', arg_type=get_enum_type(TierType), help='The desired commitment tier for this account to use.')

    with self.argument_context('dla account create') as c:
        c.argument('resource_group_name', resource_group_name_type, validator=None)
        c.argument('account_name', datalake_analytics_name_type, options_list=('--account', '-n'), completer=None)

    with self.argument_context('dla account update') as c:
        c.argument('firewall_state', help='Enable/disable existing firewall rules.', arg_type=get_enum_type(FirewallState))
        c.argument('allow_azure_ips', help='Allow/block Azure originating IPs through the firewall', arg_type=get_enum_type(FirewallAllowAzureIpsState))
        c.argument('max_job_count', help='The maximum supported jobs running under the account at the same time.', type=int)
        c.argument('max_degree_of_parallelism', help='The maximum supported degree of parallelism for this account.', type=int)
        c.argument('query_store_retention', help='The number of days that job metadata is retained.', type=int)

    with self.argument_context('dla account list') as c:
        c.argument('resource_group_name', resource_group_name_type, validator=None)

    # storage
    with self.argument_context('dla account blob-storage') as c:
        c.argument('access_key', help='the access key associated with this Azure Storage account that will be used to connect to it')
        c.argument('suffix', help='the optional suffix for the storage account')
        c.argument('storage_account_name', help='Name of an existing storage account to link to.')

    # job
    with self.argument_context('dla job submit') as c:
        c.argument('compile_mode', arg_type=get_enum_type(CompileMode), help='Indicates the type of compilation to be done on this job. Valid values are: \'Semantic\' (Only performs semantic checks and necessary sanity checks), \'Full\' (full compilation) and \'SingleBox\' (Full compilation performed locally)')
        c.argument('compile_only', help='Indicates that the submission should only build the job and not execute if set to true.', action='store_true')
        c.argument('script', completer=FilesCompleter(), help="The script to submit. This is either the script contents or use `@<file path>` to load the script from a file")
        c.argument('pipeline_id', help='Job relationship pipeline GUID.')
        c.argument('pipeline_name', help='Friendly name of the job relationship pipeline.')
        c.argument('pipeline_uri', help='Unique pipeline URI which links to the originating service for this pipeline.')
        c.argument('run_id', help='GUID of the iteration of this pipeline.')
        c.argument('recurrence_id', help='Recurrence GUID, unique per activity/script, regardless of iteration. Links different occurrences of the same job together.')
        c.argument('recurrence_name', help='Friendly recurrence nae for the correlation between jobs.')

    with self.argument_context('dla job wait') as c:
        c.argument('max_wait_time_sec', help='The maximum amount of time to wait before erroring out. Default value is to never timeout. Any value <= 0 means never timeout', type=int)
        c.argument('wait_interval_sec', help='The polling interval between checks for the job status, in seconds.', type=int)

    with self.argument_context('dla job list') as c:
        c.argument('submitted_after', help='A filter which returns jobs only submitted after the specified time, in ISO-8601 format.', type=datetime_format)
        c.argument('submitted_before', help='A filter which returns jobs only submitted before the specified time, in ISO-8601 format.', type=datetime_format)
        c.argument('state', arg_type=get_enum_type(JobState), help='A filter which returns jobs with only the specified state(s).', nargs='*')
        c.argument('result', arg_type=get_enum_type(JobResult), help='A filter which returns jobs with only the specified result(s).', nargs='*')
        c.argument('submitter', help='A filter which returns jobs only by the specified submitter.')
        c.argument('name', help='A filter which returns jobs only by the specified friendly name.')
        c.argument('pipeline_id', help='A filter which returns jobs only containing the specified pipeline_id.')
        c.argument('recurrence_id', help='A filter which returns jobs only containing the specified recurrence_id.')

    # credential
    with self.argument_context('dla catalog credential') as c:
        c.argument('uri', help='URI of the external data source.')

    with self.argument_context('dla catalog credential create') as c:
        c.argument('credential_user_password', options_list=['--password', '-p'], help='Password for the credential user. Will prompt if not given.')
        c.argument('credential_user_name', options_list=['--user-name'])

    with self.argument_context('dla catalog credential update') as c:
        c.argument('credential_user_name', options_list=['--user-name'])
        c.argument('credential_user_password', options_list=['--password', '-p'], help='Current password for the credential user. Will prompt if not given.')
        c.argument('new_credential_user_password', options_list=['--new-password'], help='New password for the credential user. Will prompt if not given.')

    # compute policy
    with self.argument_context('dla account compute_policy') as c:
        c.argument('max_dop_per_job', help='The maximum degree of parallelism allowed per job for this policy. At least one of --min-priority-per-job and --max-dop-per-job must be specified.', type=int)
        c.argument('min_priority_per_job', help='The minimum priority allowed per job for this policy. At least one of --min-priority-per-job and --max-dop-per-job must be specified.', type=int)

    with self.argument_context('dla account compute_policy create') as c:
        c.argument('object_id', help='The Azure Active Directory object ID of the user, group or service principal to apply the policy to.')
        c.argument('object_type', arg_type=get_enum_type(AADObjectType), help='The Azure Active Directory object type associated with the supplied object id.')



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

import array

from azure.cli.command_modules.keyvault.security_domain.mod_math import ModMath


class Share:
    def __init__(self, x, v):
        self.x = x
        self.v = v

    @staticmethod
    def from_uint16(w):
        x = w >> 9
        v = w & 0x1ff
        return Share(x, v)

    def to_uint16(self):
        return (self.x << 8) | self.v


class ByteShares:
    def __init__(self, required, secret_byte):
        self.coefficients = ByteShares.init_coefficients(required, secret_byte)

    @staticmethod
    def init_coefficients(required, secret_byte):
        coefficients = array.array('H')
        for _ in range(required - 1):
            coefficients.append(ModMath.get_random())
        coefficients.append(secret_byte)
        return coefficients

    def set_secret_byte(self, secret_byte):
        self.coefficients[-1] = secret_byte

    def make_share(self, x):
        v = ModMath.multiply(self.coefficients[0], x)
        v = ModMath.add(v, self.coefficients[1])

        for i in range(2, len(self.coefficients)):
            v = ModMath.multiply(v, x)
            v = ModMath.add(v, self.coefficients[i])
        return Share(x, v)

    @staticmethod
    def get_secret(shares, required):
        secret = 0
        for i in range(required):
            numerator = denominator = 1
            si = Share.from_uint16(shares[i])
            for j in range(required):
                if i == j:
                    continue
                sj = Share.from_uint16(shares[j])
                numerator = ModMath.multiply(numerator, sj.x)
                diff = ModMath.subtract(sj.x, si.x)
                denominator = ModMath.multiply(diff, denominator)

            invert = ModMath.invert(denominator)
            ci = ModMath.multiply(numerator, invert)
            tmp = ModMath.multiply(ci, si.v)
            secret = ModMath.add(secret, tmp)

        return secret



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.log import get_logger
from knack.util import CLIError
from azure.cli.command_modules.redis._client_factory import cf_redis

logger = get_logger(__name__)

allowed_c_family_sizes = ['c0', 'c1', 'c2', 'c3', 'c4', 'c5', 'c6']
allowed_p_family_sizes = ['p1', 'p2', 'p3', 'p4', 'p5']
wrong_vmsize_error = CLIError('Invalid VM size. Example for Valid values: '
                              'For Standard Sku : (C0, C1, C2, C3, C4, C5, C6), '
                              'for Premium Sku : (P1, P2, P3, P4, P5)')

# region Custom Commands


# pylint: disable=unused-argument
def cli_redis_export(cmd, client, resource_group_name, name, prefix, container, file_format=None):
    from azure.mgmt.redis.models import ExportRDBParameters
    parameters = ExportRDBParameters(prefix=prefix, container=container, format=file_format)
    return client.export_data(resource_group_name, name, parameters)


# pylint: disable=unused-argument
def cli_redis_update(cmd, instance, sku=None, vm_size=None):
    from azure.mgmt.redis.models import RedisUpdateParameters
    if sku is not None:
        instance.sku.name = sku

    if vm_size is not None:
        instance.sku.family = vm_size[0]
        instance.sku.capacity = vm_size[1:]

    # avoid setting memory configs for basic sku
    if instance.sku.name == 'Basic':
        memory_configs = ['maxmemory-reserved', 'maxfragmentationmemory-reserved', 'maxmemory-delta']
        for memory_config in memory_configs:
            if memory_config in instance.redis_configuration:
                instance.redis_configuration.pop(memory_config)

    # trim RDB and AOF connection strings
    rdb_aof_connection_strings = ['rdb-storage-connection-string',
                                  'aof-storage-connection-string-0',
                                  'aof-storage-connection-string-1']
    for connection_string in rdb_aof_connection_strings:
        if connection_string in instance.redis_configuration:
            instance.redis_configuration.pop(connection_string)

    # Trim zonal-configuration
    if 'zonal-configuration' in instance.redis_configuration:
        instance.redis_configuration.pop('zonal-configuration')

    # pylint: disable=too-many-function-args
    update_params = RedisUpdateParameters(
        redis_configuration=instance.redis_configuration,
        enable_non_ssl_port=instance.enable_non_ssl_port,
        tenant_settings=instance.tenant_settings,
        shard_count=instance.shard_count,
        minimum_tls_version=instance.minimum_tls_version,
        sku=instance.sku,
        tags=instance.tags
    )
    return update_params


# pylint: disable=unused-argument
def cli_redis_create(cmd, client,
                     resource_group_name, name, location, sku, vm_size, tags=None,
                     redis_configuration=None, enable_non_ssl_port=None, tenant_settings=None,
                     shard_count=None, minimum_tls_version=None, subnet_id=None, static_ip=None,
                     zones=None, replicas_per_master=None):
    # pylint:disable=line-too-long
    if ((sku.lower() in ['standard', 'basic'] and vm_size.lower() not in allowed_c_family_sizes) or (sku.lower() in ['premium'] and vm_size.lower() not in allowed_p_family_sizes)):
        raise wrong_vmsize_error
    tenant_settings_in_json = {}
    if tenant_settings is not None:
        for item in tenant_settings:
            tenant_settings_in_json.update(get_key_value_pair(item))
    from azure.mgmt.redis.models import RedisCreateParameters, Sku
    # pylint: disable=too-many-function-args
    params = RedisCreateParameters(
        sku=Sku(name=sku, family=vm_size[0], capacity=vm_size[1:]),
        location=location,
        redis_configuration=redis_configuration,
        enable_non_ssl_port=enable_non_ssl_port,
        replicas_per_master=replicas_per_master,
        tenant_settings=tenant_settings_in_json,
        shard_count=shard_count,
        minimum_tls_version=minimum_tls_version,
        subnet_id=subnet_id,
        static_ip=static_ip,
        zones=zones,
        tags=tags)
    return client.create(resource_group_name, name, params)


def get_key_value_pair(string):
    result = {}
    if string:
        kvp = string.split('=', 1)
        result = {kvp[0]: kvp[1]} if len(kvp) > 1 else {string: ''}
    return result


# pylint: disable=unused-argument
def cli_redis_create_server_link(cmd, client, resource_group_name, name, server_to_link, replication_role):
    redis_client = cf_redis(cmd.cli_ctx)
    from azure.cli.core.commands.client_factory import get_subscription_id
    from msrestazure.tools import is_valid_resource_id, resource_id
    if not is_valid_resource_id(server_to_link):
        server_to_link = resource_id(
            subscription=get_subscription_id(cmd.cli_ctx),
            resource_group=resource_group_name,
            namespace='Microsoft.Cache', type='Redis',
            name=server_to_link
        )

    cache_to_link = get_cache_from_resource_id(redis_client, server_to_link)

    from azure.mgmt.redis.models import RedisLinkedServerCreateParameters
    params = RedisLinkedServerCreateParameters(linked_redis_cache_id=cache_to_link.id,
                                               linked_redis_cache_location=cache_to_link.location,
                                               server_role=replication_role)
    return client.create(resource_group_name, name, cache_to_link.name, params)


def cli_redis_list_cache(client, resource_group_name=None):
    cache_list = client.list_by_resource_group(resource_group_name=resource_group_name) \
        if resource_group_name else client.list()
    return list(cache_list)


def get_cache_from_resource_id(client, cache_resource_id):
    from msrestazure.tools import parse_resource_id
    id_comps = parse_resource_id(cache_resource_id)
    return client.get(id_comps['resource_group'], id_comps['name'])

# endregion



# coding=utf-8
# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.help_files import helps  # pylint: disable=unused-import
# pylint: disable=line-too-long, too-many-lines

helps['signalr'] = """
type: group
short-summary: Manage Azure SignalR Service.
"""

helps['signalr cors'] = """
type: group
short-summary: Manage CORS for Azure SignalR Service.
"""

helps['signalr network-rule'] = """
type: group
short-summary: Manage network rules.
"""

helps['signalr upstream'] = """
type: group
short-summary: Manage upstream settings.
"""

helps['signalr cors add'] = """
type: command
short-summary: Add allowed origins to a SignalR Service
examples:
  - name: Add a list of allowed origins to a SignalR Service
    text: >
        az signalr cors add -n MySignalR -g MyResourceGroup --allowed-origins "http://example1.com" "https://example2.com"
"""

helps['signalr cors list'] = """
type: command
short-summary: List allowed origins of a SignalR Service
"""

helps['signalr cors remove'] = """
type: command
short-summary: Remove allowed origins from a SignalR Service
examples:
  - name: Remove a list of allowed origins from a SignalR Service
    text: >
        az signalr cors remove -n MySignalR -g MyResourceGroup --allowed-origins "http://example1.com" "https://example2.com"
"""

helps['signalr create'] = """
type: command
short-summary: Creates a SignalR Service.
examples:
  - name: Create a SignalR Service with the Standard SKU and serverless mode and enable messaging logs.
    text: >
        az signalr create -n MySignalR -g MyResourceGroup --sku Standard_S1 --unit-count 1 --service-mode Serverless --enable-message-logs True
"""

helps['signalr delete'] = """
type: command
short-summary: Deletes a SignalR Service.
examples:
  - name: Delete a SignalR Service.
    text: >
        az signalr delete -n MySignalR -g MyResourceGroup
"""

helps['signalr key'] = """
type: group
short-summary: Manage keys for Azure SignalR Service.
"""

helps['signalr key list'] = """
type: command
short-summary: List the access keys for a SignalR Service.
examples:
  - name: Get the primary key for a SignalR Service.
    text: >
        az signalr key list -n MySignalR -g MyResourceGroup --query primaryKey -o tsv
"""

helps['signalr key renew'] = """
type: command
short-summary: Regenerate the access key for a SignalR Service.
examples:
  - name: Renew the secondary key for a SignalR Service.
    text: >
        az signalr key renew -n MySignalR -g MyResourceGroup --key-type secondary
"""

helps['signalr list'] = """
type: command
short-summary: Lists all the SignalR Service under the current subscription.
examples:
  - name: List SignalR Service and show the results in a table.
    text: >
        az signalr list -o table
  - name: List SignalR Service in a resource group and show the results in a table.
    text: >
        az signalr list -g MySignalR -o table
"""

helps['signalr restart'] = """
type: command
short-summary: Restart an existing SignalR Service.
examples:
  - name: Restart a SignalR Service instance.
    text: >
        az signalr restart -n MySignalR -g MyResourceGroup
"""

helps['signalr show'] = """
type: command
short-summary: Get the details of a SignalR Service.
examples:
  - name: Get the sku for a SignalR Service.
    text: >
        az signalr show -n MySignalR -g MyResourceGroup --query sku
"""

helps['signalr update'] = """
type: command
short-summary: Update an existing SignalR Service.
examples:
  - name: Update unit count to scale the service.
    text: >
        az signalr update -n MySignalR -g MyResourceGroup --sku Standard_S1 --unit-count 50
  - name: Update service mode.
    text: >
        az signalr update -n MySignalR -g MyResourceGroup --service-mode Serverless
  - name: Update for enabling messaging logs in the service.
    text: >
        az signalr update -n MySignalR -g MyResourceGroup --enable-message-logs True
"""

helps['signalr upstream list'] = """
type: command
short-summary: List upstream settings of an existing SignalR Service.
"""

helps['signalr upstream update'] = """
type: command
short-summary: Update order sensitive upstream settings for an existing SignalR Service.
examples:
  - name: Update two upstream settings to handle messages and connections separately.
    text: >
        az signalr upstream update -n MySignalR -g MyResourceGroup --template url-template="http://host-connections.com" category-pattern="connections" --template url-template="http://host-connections.com" category-pattern="messages"
  - name: Update one upstream setting to handle a specific event in a specific hub.
    text: >
        az signalr upstream update -n MySignalR -g MyResourceGroup --template url-template="http://host.com/{hub}/{event}/{category}" category-pattern="messages" event-pattern="broadcast" hub-pattern="chat"
"""

helps['signalr upstream clear'] = """
type: command
short-summary: List upstream settings of an existing SignalR Service.
"""

helps['signalr network-rule list'] = """
type: command
short-summary: Get the Network access control of SignalR Service.
"""

helps['signalr network-rule update'] = """
type: command
short-summary: Update the Network access control of SignalR Service.
examples:
  - name: Set allowing RESTAPI only for public network.
    text: >
        az signalr network-rule update --public-network -n MySignalR -g MyResourceGroup --allow RESTAPI
  - name: Set allowing client connection and server connection for a private endpoint connection
    text: >
        az signalr network-rule update --connection-name MyPrivateEndpointConnection -n MySignalR -g MyResourceGroup --allow ClientConnection ServerConnection
  - name: Set denying client connection for both public network and private endpoint connections
    text: >
        az signalr network-rule update --public-network --connection-name MyPrivateEndpointConnection1 MyPrivateEndpointConnection2 -n MySignalR -g MyResourceGroup --deny ClientConnection
"""



# --------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#
# Code generated by Microsoft (R) AutoRest Code Generator.
# Changes may cause incorrect behavior and will be lost if the code is
# regenerated.
# --------------------------------------------------------------------------
# pylint: disable=too-many-lines
# pylint: disable=too-many-statements

from azure.cli.core.commands.parameters import (
    get_three_state_flag,
    get_enum_type
)
from azure.cli.core.commands.validators import validate_file_or_dict
from ..action import (
    AddSoldTo,
    AddEnabledAzurePlans,
    AddLabels
)


def load_arguments(self, _):

    with self.argument_context('billing account list') as c:
        c.argument('expand', type=str,
                   help='May be used to expand the soldTo, invoice sections and billing profiles.')

    with self.argument_context('billing account show') as c:
        c.argument('account_name', options_list=['--name', '-n'], type=str, help='The ID that uniquely identifies a '
                   'billing account.')
        c.argument('expand', type=str,
                   help='May be used to expand the soldTo, invoice sections and billing profiles.')

    with self.argument_context('billing account update') as c:
        c.argument('account_name', options_list=['--name', '-n'], type=str, help='The ID that uniquely identifies a '
                   'billing account.')
        c.argument('display_name', type=str, help='The billing account name.')
        c.argument('sold_to', action=AddSoldTo, nargs='*', help='The address of the individual or organization that is '
                   'responsible for the billing account.')
        c.argument('departments', type=validate_file_or_dict, help='The departments associated to the enrollment. '
                   'Expected value: json-string/@json-file.')
        c.argument('enrollment_accounts', type=validate_file_or_dict, help='The accounts associated to the enrollment. '
                   'Expected value: json-string/@json-file.')
        c.argument('billing_profiles_value', type=validate_file_or_dict, help='The billing profiles associated with '
                   'the billing account. Expected value: json-string/@json-file.')

    with self.argument_context('billing account wait') as c:
        c.argument('account_name', options_list=['--name', '-n'], type=str, help='The ID that uniquely identifies a '
                   'billing account.')
        c.argument('expand', type=str,
                   help='May be used to expand the soldTo, invoice sections and billing profiles.')

    with self.argument_context('billing balance show') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')

    with self.argument_context('billing instruction list') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')

    with self.argument_context('billing instruction show') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('name', options_list=['--name', '-n'], type=str, help='Instruction Name.')

    with self.argument_context('billing instruction create') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('name', options_list=['--name', '-n'], type=str, help='Instruction Name.')
        c.argument('amount', type=float, help='The amount budgeted for this billing instruction.')
        c.argument('start_date', help='The date this billing instruction goes into effect.')
        c.argument('end_date', help='The date this billing instruction is no longer in effect.')
        c.argument('creation_date', help='The date this billing instruction was created.')

    with self.argument_context('billing profile list') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('expand', type=str, help='May be used to expand the invoice sections.')

    with self.argument_context('billing profile show') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', options_list=['--name', '-n'], type=str, help='The ID that uniquely identifies a '
                   'billing profile.')
        c.argument('expand', type=str, help='May be used to expand the invoice sections.')

    with self.argument_context('billing profile create') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', options_list=['--name', '-n'], type=str, help='The ID that uniquely identifies a '
                   'billing profile.')
        c.argument('display_name', type=str, help='The name of the billing profile.')
        c.argument('po_number', type=str, help='The purchase order name that will appear on the invoices generated for '
                   'the billing profile.')
        c.argument('bill_to', action=AddSoldTo, nargs='*', help='Billing address.')
        c.argument('invoice_email_opt_in', arg_type=get_three_state_flag(), help='Flag controlling whether the '
                   'invoices for the billing profile are sent through email.')
        c.argument('enabled_azure_plans', action=AddEnabledAzurePlans, nargs='*', help='Information about the enabled '
                   'azure plans.')
        c.argument('invoice_sections_value', type=validate_file_or_dict, help='The invoice sections associated to the '
                   'billing profile. Expected value: json-string/@json-file.')

    with self.argument_context('billing profile update') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', options_list=['--name', '-n'], type=str, help='The ID that uniquely identifies a '
                   'billing profile.')
        c.argument('display_name', type=str, help='The name of the billing profile.')
        c.argument('po_number', type=str, help='The purchase order name that will appear on the invoices generated for '
                   'the billing profile.')
        c.argument('bill_to', action=AddSoldTo, nargs='*', help='Billing address.')
        c.argument('invoice_email_opt_in', arg_type=get_three_state_flag(), help='Flag controlling whether the '
                   'invoices for the billing profile are sent through email.')
        c.argument('enabled_azure_plans', action=AddEnabledAzurePlans, nargs='*', help='Information about the enabled '
                   'azure plans.')
        c.argument('invoice_sections_value', type=validate_file_or_dict, help='The invoice sections associated to the '
                   'billing profile. Expected value: json-string/@json-file.')

    with self.argument_context('billing profile wait') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', options_list=['--name', '-n'], type=str, help='The ID that uniquely identifies a '
                   'billing profile.')
        c.argument('expand', type=str, help='May be used to expand the invoice sections.')

    with self.argument_context('billing customer list') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('search', type=str, help='Used for searching customers by their name. Any customer with name '
                   'containing the search text will be included in the response')
        c.argument('filter_', options_list=['--filter'], type=str,
                   help='May be used to filter the list of customers.')

    with self.argument_context('billing customer show') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('customer_name', options_list=['--name', '-n', '--customer-name'], type=str, help='The ID that '
                   'uniquely identifies a customer.')
        c.argument('expand', type=str, help='May be used to expand enabledAzurePlans and resellers')

    with self.argument_context('billing invoice section list') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')

    with self.argument_context('billing invoice section show') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('invoice_section_name', options_list=['--name', '-n', '--invoice-section-name'], type=str, help=''
                   'The ID that uniquely identifies an invoice section.')

    with self.argument_context('billing invoice section create') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('invoice_section_name', options_list=['--name', '-n', '--invoice-section-name'], type=str, help=''
                   'The ID that uniquely identifies an invoice section.')
        c.argument('display_name', type=str, help='The name of the invoice section.')
        c.argument('labels', action=AddLabels, nargs='*', help='Dictionary of metadata associated with the invoice '
                   'section. Expect value: KEY1=VALUE1 KEY2=VALUE2 ...')

    with self.argument_context('billing invoice section update') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('invoice_section_name', options_list=['--name', '-n', '--invoice-section-name'], type=str, help=''
                   'The ID that uniquely identifies an invoice section.')
        c.argument('display_name', type=str, help='The name of the invoice section.')
        c.argument('labels', action=AddLabels, nargs='*', help='Dictionary of metadata associated with the invoice '
                   'section. Expect value: KEY1=VALUE1 KEY2=VALUE2 ...')

    with self.argument_context('billing invoice section wait') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('invoice_section_name', options_list=['--name', '-n', '--invoice-section-name'], type=str, help=''
                   'The ID that uniquely identifies an invoice section.')

    with self.argument_context('billing permission list') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('invoice_section_name', type=str, help='The ID that uniquely identifies an invoice section.')
        c.argument('customer_name', type=str, help='The ID that uniquely identifies a customer.')

    with self.argument_context('billing subscription list') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('invoice_section_name', type=str, help='The ID that uniquely identifies an invoice section.')
        c.argument('customer_name', type=str, help='The ID that uniquely identifies a customer.')

    with self.argument_context('billing subscription show') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')

    with self.argument_context('billing subscription update') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('subscription_billing_status', arg_type=get_enum_type(['Active', 'Inactive', 'Abandoned', 'Deleted',
                                                                          'Warning']), help='The current billing '
                   'status of the subscription.')
        c.argument('cost_center', type=str, help='The cost center applied to the subscription.')
        c.argument('sku_id', type=str, help='The sku ID of the Azure plan for the subscription.')

    with self.argument_context('billing subscription move') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('destination_invoice_section_id', type=str, help='The destination invoice section id.')

    with self.argument_context('billing subscription validate-move') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('destination_invoice_section_id', type=str, help='The destination invoice section id.')

    with self.argument_context('billing subscription wait') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')

    with self.argument_context('billing product list') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('invoice_section_name', type=str, help='The ID that uniquely identifies an invoice section.')
        c.argument('filter_', options_list=['--filter'], type=str, help='May be used to filter by product type. The '
                   'filter supports \'eq\', \'lt\', \'gt\', \'le\', \'ge\', and \'and\'. It does not currently support '
                   '\'ne\', \'or\', or \'not\'. Tag filter is a key value pair string where key and value are '
                   'separated by a colon (:).')
        c.argument('customer_name', type=str, help='The ID that uniquely identifies a customer.')

    with self.argument_context('billing product show') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('product_name', options_list=['--name', '-n', '--product-name'], type=str, help='The ID that '
                   'uniquely identifies a product.')

    with self.argument_context('billing product update') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('product_name', options_list=['--name', '-n', '--product-name'], type=str, help='The ID that '
                   'uniquely identifies a product.')
        c.argument('auto_renew', arg_type=get_enum_type(['Off', 'On']), help='Indicates whether auto renewal is turned '
                   'on or off for a product.')
        c.argument('status', arg_type=get_enum_type(['Active', 'Inactive', 'PastDue', 'Expiring', 'Expired',
                                                     'Disabled', 'Cancelled', 'AutoRenew']), help='The current status '
                   'of the product.')
        c.argument('billing_frequency', arg_type=get_enum_type(['OneTime', 'Monthly', 'UsageBased']), help='The '
                   'frequency at which the product will be billed.')

    with self.argument_context('billing product move') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('product_name', options_list=['--name', '-n', '--product-name'], type=str, help='The ID that '
                   'uniquely identifies a product.')
        c.argument('destination_invoice_section_id', type=str, help='The destination invoice section id.')

    with self.argument_context('billing product validate-move') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('product_name', options_list=['--name', '-n', '--product-name'], type=str, help='The ID that '
                   'uniquely identifies a product.')
        c.argument('destination_invoice_section_id', type=str, help='The destination invoice section id.')

    with self.argument_context('billing invoice list') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('period_start_date', type=str, help='The start date to fetch the invoices. The date should be '
                   'specified in MM-DD-YYYY format.')
        c.argument('period_end_date', type=str, help='The end date to fetch the invoices. The date should be specified '
                   'in MM-DD-YYYY format.')

    with self.argument_context('billing invoice show') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('name', options_list=['--name', '-n'], type=str,
                   help='The ID that uniquely identifies an invoice.')

    with self.argument_context('billing transaction list') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('invoice_name', type=str, help='The ID that uniquely identifies an invoice.')

    with self.argument_context('billing policy update') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('marketplace_purchases', arg_type=get_enum_type(['AllAllowed', 'OnlyFreeAllowed', 'NotAllowed']),
                   help='The policy that controls whether Azure marketplace purchases are allowed for a billing '
                   'profile.')
        c.argument('reservation_purchases', arg_type=get_enum_type(['Allowed', 'NotAllowed']), help='The policy that '
                   'controls whether Azure reservation purchases are allowed for a billing profile.')
        c.argument('view_charges', arg_type=get_enum_type(['Allowed', 'NotAllowed']), help='The policy that controls '
                   'whether users with Azure RBAC access to a subscription can view its charges.')

    with self.argument_context('billing property update') as c:
        c.argument('cost_center', type=str, help='The cost center applied to the subscription.')

    with self.argument_context('billing role-definition list') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('invoice_section_name', type=str, help='The ID that uniquely identifies an invoice section.')

    with self.argument_context('billing role-assignment list') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('invoice_section_name', type=str, help='The ID that uniquely identifies an invoice section.')

    with self.argument_context('billing role-assignment delete') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('profile_name', type=str, help='The ID that uniquely identifies a billing profile.')
        c.argument('invoice_section_name', type=str, help='The ID that uniquely identifies an invoice section.')
        c.argument('name', options_list=['--name', '-n'], type=str, help='The ID that uniquely identifies a role '
                   'assignment.')

    with self.argument_context('billing agreement list') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('expand', type=str, help='May be used to expand the participants.')

    with self.argument_context('billing agreement show') as c:
        c.argument('account_name', type=str, help='The ID that uniquely identifies a billing account.')
        c.argument('name', options_list=['--name', '-n'], type=str, help='The ID that uniquely identifies an '
                   'agreement.')
        c.argument('expand', type=str, help='May be used to expand the participants.')
